/**************************************************************************/
/*                                                                        */
/*  Thibaut Balabonski, Sylvain Conchon, Jean-Christophe Filliâtre,       */
/*  Kim Nguyen, Laurent Sartre                                            */
/*                                                                        */
/*  Informatique - MP2I/MPI - CPGE 1re et 2e années.                      */
/*  Cours et exercices corrigés. Éditions Ellipses, 2022.                 */
/*                                                                        */
/*  https://www.informatique-mpi.fr/                                      */
/*                                                                        */
/**************************************************************************/

/* Programmation dynamique

   Rendu de monnaie : étant donné un système monétaire `coins`,
   par exemple `coins = { 1, 2, 5, 10, 20, 50 }`, quel est le plus
   petit nombre de billets et de pièces pour former la somme `s` ?

*/

#include "prelude.h"
#include "dp_change.h"

// temps et espace O(ns)
int dp_change(int coins[], int n, int s) {
  assert(n > 0 && coins[0] == 1 && s >= 0);
  int **m = calloc(n, sizeof(int*));
  for (int i = 0; i < n; i++) m[i] = calloc(s + 1, sizeof(int));
  // m[i][t] = solution pour une somme t avec les pièces 0,1,...,i
  for (int t = 1; t <= s; t++)
    m[0][t] = t;
  for (int i = 1; i < n; i++)
    for (int t = 1; t <= s; t++) {
      m[i][t] = m[i-1][t]; // sans coins[i]
      int d = t - coins[i];
      if (d >= 0 && m[i][d] < m[i][t]) m[i][t] = 1 + m[i][d]; // avec
    }
  int r = m[n-1][s];
  for (int i = 0; i < n; i++) free(m[i]);
  free(m);
  return r;
}

// en ne conservant qu'une seule ligne => espace O(s) seulement
int dp_change2(int coins[], int n, int s) {
  assert(n > 0 && coins[0] == 1 && s >= 0);
  int *m = calloc(s + 1, sizeof(int));
  for (int t = 1; t <= s; t++)
    m[t] = t;
  for (int i = 1; i < n; i++)
    for (int t = 1; t <= s; t++) {
      int d = t - coins[i];
      if (d >= 0 && m[d] < m[t]) m[t] = 1 + m[d];
    }
  int r = m[s];
  free(m);
  return r;
}
